/* sd-card-helper: sd card helper daemon to access sd card storage
 * Copyright (C) 2007-2011 David Zeuthen <zeuthen@gmail.com>
 * Copyright (C) 2018 Robert Bosch Power Tools GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef __SD_CARD_HELPER_LOGGING_H__
#define __SD_CARD_HELPER_LOGGING_H__

#include <gio/gio.h>
G_BEGIN_DECLS

/**
 * sd_card_helperLogLevel:
 * @SD_CARD_HELPER_LOG_LEVEL_DEBUG: Debug messages.
 * @SD_CARD_HELPER_LOG_LEVEL_INFO: Informational messages.
 * @SD_CARD_HELPER_LOG_LEVEL_NOTICE: Messages that the administrator should
 * take notice of.
 * @SD_CARD_HELPER_LOG_LEVEL_WARNING: Warning messages.
 * @SD_CARD_HELPER_LOG_LEVEL_ERROR: Error messages.
 *
 * Logging levels. The level @SD_CARD_HELPER_LOG_LEVEL_NOTICE and above goes to
 * journal.
 *
 * Unlike g_warning() and g_error(), none of these logging levels causes the
 * program to ever terminate.
 */
typedef enum
{
  SD_CARD_HELPER_LOG_LEVEL_DEBUG    = G_LOG_LEVEL_DEBUG,
  SD_CARD_HELPER_LOG_LEVEL_INFO     = G_LOG_LEVEL_INFO,
  SD_CARD_HELPER_LOG_LEVEL_MESSAGE  = G_LOG_LEVEL_MESSAGE,
  SD_CARD_HELPER_LOG_LEVEL_WARNING  = G_LOG_LEVEL_WARNING,
  SD_CARD_HELPER_LOG_LEVEL_CRITICAL = G_LOG_LEVEL_CRITICAL,
  SD_CARD_HELPER_LOG_LEVEL_ERROR    = G_LOG_LEVEL_ERROR
} SdCardHelperLogLevel;

void sd_card_helper_log (SdCardHelperLogLevel level, const gchar *function,
                         const gchar *location, const gchar *format, ...)
    G_GNUC_PRINTF (4, 5);

/**
 * sd_card_helper_critical:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %sd_card_helper_LOG_LEVEL_DEBUG.
 * Warning: ONLY WORKS WHEN COMPILED WITH --enable-debug
 *
 * See #sd_card_helperLogLevel for more details.
 */
#ifdef DEBUG
#define sd_card_helper_debug(args...)                                         \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_DEBUG, G_STRFUNC, G_STRLOC,    \
                      args)
#else
#define sd_card_helper_debug(args...)
#endif

/**
 * sd_card_helper_info:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_INFO.
 * Warning: ONLY WORKS WHEN COMPILED WITH --enable-debug
 *
 * See #sd_card_helperLogLevel for more details.
 */
#ifdef DEBUG
#define sd_card_helper_info(args...)                                          \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_INFO, G_STRFUNC, G_STRLOC, args)
#else
#define sd_card_helper_info(args...)
#endif

/**
 * sd_card_helper_notice:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_MESSAGE.
 *
 * See #sd_card_helperLogLevel for more details.
 */
#define sd_card_helper_notice(args...)                                        \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_MESSAGE, G_STRFUNC, G_STRLOC,  \
                      args)

/**
 * sd_card_helper_message:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_MESSAGE.
 *
 * See #sd_card_helperLogLevel for more details.
 */
#define sd_card_helper_message(args...)                                       \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_MESSAGE, G_STRFUNC, G_STRLOC,  \
                      args)

/**
 * sd_card_helper_warning:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_WARNING.
 *
 * See #sd_card_helperLogLevel for more details.
 */
#define sd_card_helper_warning(args...)                                       \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_WARNING, G_STRFUNC, G_STRLOC,  \
                      args)

/**
 * sd_card_helper_critical:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_CRITICAL.
 *
 * See #sd_card_helperLogLevel for more details.
 */
#define sd_card_helper_critical(args...)                                      \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_CRITICAL, G_STRFUNC, G_STRLOC, \
                      args)

/**
 * sd_card_helper_error:
 * @args...: printf()-style format string and arguments
 *
 * Logging macro for %SD_CARD_HELPER_LOG_LEVEL_ERROR.
 * Warning: THIS ALSO TERMINATES THE PROCESS with the abort() call!
 *
 * See #sd_card_helperLogLevel for more details.
 */
#define sd_card_helper_error(args...)                                         \
  sd_card_helper_log (SD_CARD_HELPER_LOG_LEVEL_ERROR, G_STRFUNC, G_STRLOC,    \
                      args)

G_END_DECLS

#endif /* __SD_CARD_HELPER_LOGGING_H__ */
