/* sd-card-helper: sd card helper daemon to access sd card storage
 * Copyright (C) 2018 Robert Bosch Power Tools GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Author: Vinay Kumar Shettar <vinaykumar.shettar@in.bosch.com>
 */

#include <blkid/blkid.h>
#include <blockdev/blockdev.h>
#include <blockdev/fs.h>
#include <libfdisk.h>
#include <libudev.h>
#include <mntent.h>
#include <sd-card-helper-logging.h>
#include <sd-card-helper.h>

#define UDEV_ADD_FILTER "add"
#define UDEV_REMOVE_FILTER "remove"
#define UDEV_SUBSYSTEM_FILTER "block"
#define UDEV_DEVTYPE_FILTER "disk"
#define MMC_BLOCK "mmcblk0"
#define DEVICE_MMC "/dev/mmcblk0"
#define DEVICE_MMC_PARTITION_NAME "/dev/mmcblk0p1"
#define SD_CARD_MOUNT_DIR "/media/SD_CARD"

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief setup mount directory for mount stroge partition in device.
 * @return TRUE on success of setup of mount dir FALSE otherwises.
 */
static gboolean
setup_mount_dir (const gchar *filename)
{
  if (NULL == filename)
    {
      sd_card_helper_critical ("NULL filename\n");
      return FALSE;
    }
  if (!g_file_test (filename, G_FILE_TEST_IS_DIR))
    {
      int ret;
      ret = g_mkdir_with_parents (filename, 0766);

      if (ret != 0)
        {
          sd_card_helper_critical ("Couldn't create mount directory\n");
          return FALSE;
        }
      else
        {
          g_chmod (filename, 0766);
          sd_card_helper_debug ("Directory Created\n");
        }
    }
  else
    {
      g_chmod (filename, 0766);
      sd_card_helper_debug ("Directory present\n");
    }
  return TRUE;
}

/**
 * @author Vinay Kumar
 * @date 28 May 2020
 * @brief cleanup mount directory after unmount.
 * @return TRUE on success of cleanup of mount dir FALSE otherwises.
 */
static gboolean
cleanup_mount_dir (const gchar *filename)
{
  if (NULL == filename)
    {
      sd_card_helper_critical ("NULL filename\n");
      return FALSE;
    }
  if (g_file_test (filename, G_FILE_TEST_IS_DIR))
    {
      int ret;
      ret = g_rmdir (filename);

      if (ret != 0)
        {
          sd_card_helper_critical ("Couldn't remove mount directory\n");
          return FALSE;
        }
      else
        {
          sd_card_helper_debug ("Directory Removed\n");
        }
    }
  else
    {
      sd_card_helper_debug ("Directory Not present\n");
    }
  return TRUE;
}
/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Mount the Sd card to Device.
 * @param mount_path path to mount the sd card.
 * @return Mounted path on mount success and NULL othewise.
 */
static gchar *
mount_sd_card (SDCardHelper *sd_card_helper, const gchar *mount_path,
               const gchar *mount_option)
{
  g_autoptr (GError) error = NULL;
  gchar *fstype            = NULL;
  gchar *device_name       = DEVICE_MMC_PARTITION_NAME;
  gchar *mounted_path      = bd_fs_get_mountpoint (device_name, &error);

  if (mounted_path == NULL)
    {
      gint ret = 0;
      setup_mount_dir (SD_CARD_MOUNT_DIR);
      fstype = bd_fs_get_fstype (device_name, &error);
      if (fstype == NULL)
        {
          sd_card_helper_critical ("Error bd_fs_get_fstype:");
          if (error)
            sd_card_helper_critical ("%s (%s, %d)\n", error->message,
                                     g_quark_to_string (error->domain),
                                     error->code);
          g_clear_error (&error);
        }
      else
        {
          sd_card_helper_debug ("bd_fs_get_fstype : %s\n", fstype);
        }
      /*
       * TODO: Get the mount option from client process to set uid and gid
       *      to mount with normal user permission.
       */
      ret = bd_fs_mount (device_name, mount_path, fstype,
                         "rw,uid=1000,gid=1000", NULL, &error);
      if (!ret)
        {
          sd_card_helper_critical ("Error bd_fs_mount: %d\n", ret);
          sd_card_helper_critical ("%s (%s, %d)\n", error->message,
                                   g_quark_to_string (error->domain),
                                   error->code);
          g_clear_error (&error);
        }
      else
        {
          sd_card_helper_debug ("bd_fs_Mount Done: %d\n", ret);
          mounted_path = strdup (mount_path);
          com_bosch_pt_sd_card_helper_set_mount_point (
              sd_card_helper->sd_card_helper_object, mounted_path);
        }
    }
  else
    {
      sd_card_helper_debug ("Mount path exist: %s\n", mounted_path);
      com_bosch_pt_sd_card_helper_set_mount_point (
          sd_card_helper->sd_card_helper_object, mounted_path);
    }

  return mounted_path;
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Unmount the Mass Storage Partition to Device.
 * @return 1 on unmount success and 0 othewise.
 */
static gint
unmount_sd_card (SDCardHelper *sd_card_helper)
{
  g_autoptr (GError) error = NULL;
  gint ret                 = 0;
  gchar *device_name       = DEVICE_MMC_PARTITION_NAME;
  gchar *mount_path_exist  = bd_fs_get_mountpoint (device_name, &error);
  /*
   * Check if storage is mounted already or not with bd_fs_get_mountpoint()
   * This is needed if device connected to PC at init and unmount is called by
   * applicaton.
   */
  if (mount_path_exist != NULL)
    {
      ret = bd_fs_unmount (device_name, 1, 0, NULL, &error);
      if (!ret)
        {
          sd_card_helper_critical ("Error bd_fs_unmount: %d\n", ret);
          sd_card_helper_critical ("%s (%s, %d)\n", error->message,
                                   g_quark_to_string (error->domain),
                                   error->code);
        }
      else
        {
          sd_card_helper_debug ("bd_fs_unmount Done: %d\n", ret);
          com_bosch_pt_sd_card_helper_set_mount_point (
              sd_card_helper->sd_card_helper_object, NULL);
        }
    }
  else
    {
      sd_card_helper_debug ("mount path does not exist\n");
      com_bosch_pt_sd_card_helper_set_mount_point (
          sd_card_helper->sd_card_helper_object, NULL);
    }
  cleanup_mount_dir (SD_CARD_MOUNT_DIR);
  return ret;
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Handler for sd card added or removed signal from uvev gio channel.
 * @param source GIOChannel pointer.
 * @param f_en_condition         gio condition.
 * @param user_data       pointer to user data
 * @attention It will checked whether the device added or removed.
 * it send notifiction that the sdcard added or removed.
 */
static gboolean
handle_udev_events (GIOChannel *source, GIOCondition f_en_condition,
                    gpointer user_data)
{
  SDCardHelper *sd_card_helper = user_data;
  struct udev_device *device
      = udev_monitor_receive_device (sd_card_helper->udev_monitor);

  if (strcmp (udev_device_get_sysname (device), MMC_BLOCK) == 0)
    {
      if (strcmp (udev_device_get_action (device), UDEV_ADD_FILTER) == 0)
        {
          sd_card_helper_debug ("++Sd card Inserted++");
          com_bosch_pt_sd_card_helper_set_is_sd_card_present (
              sd_card_helper->sd_card_helper_object, TRUE);
        }

      if (strcmp (udev_device_get_action (device), UDEV_REMOVE_FILTER) == 0)
        {
          sd_card_helper_debug ("--Sd card Removed--");
          com_bosch_pt_sd_card_helper_set_is_sd_card_present (
              sd_card_helper->sd_card_helper_object, FALSE);
          if (!unmount_sd_card (sd_card_helper))
            {
              sd_card_helper_warning ("unmount_sd_card Failed");
            }
        }
    }

  return TRUE;
}
/**
 * @author Vinay
 * @date 31.08.2020
 * @brief check if the udev device sd card or not.
 * @param udev_device device pointer.
 */
static gboolean
is_device_sd_card (struct udev_device *device)
{
  gboolean retVal             = FALSE;
  const char *sd_card_sysname = udev_device_get_sysname (device);
  sd_card_helper_debug ("sd_card_sysname %s \n", sd_card_sysname);
  if (strncmp (sd_card_sysname, "mmcblk0", 7) == 0)
    {
      sd_card_helper_debug ("Device is SD \n");
      retVal = TRUE;
    }
  return retVal;
}

/**
 * @author Vinay
 * @date 31.08.2020
 * @brief check if the sd card present or not.
 * @param udev device pointer.
 */
static gboolean
is_sd_card_present (struct udev *device)
{
  gboolean retVal                        = FALSE;
  struct udev_list_entry *devices        = NULL;
  struct udev_list_entry *dev_list_entry = NULL;
  struct udev_enumerate *enumerate       = udev_enumerate_new (device);
  udev_enumerate_add_match_subsystem (enumerate, UDEV_SUBSYSTEM_FILTER);
  udev_enumerate_scan_devices (enumerate);
  devices = udev_enumerate_get_list_entry (enumerate);

  udev_list_entry_foreach (dev_list_entry, devices)
  {
    struct udev_device *dev;
    const char *dev_path = udev_list_entry_get_name (dev_list_entry);
    sd_card_helper_debug ("dev_path %s", dev_path);
    dev = udev_device_new_from_syspath (device, dev_path);

    if (TRUE == is_device_sd_card (dev))
      {
        retVal = TRUE;
        udev_device_unref (dev);
        break;
      }

    udev_device_unref (dev);
  }
  udev_enumerate_unref (enumerate);

  return retVal;
}

/**
 * @author Vinay
 * @date 31.08.2020
 * @brief Start monitor on sd card added or removed.
 * @param sd_card_helper SDCardHelper pointer.
 */
static void
start_udev_monitor (SDCardHelper *sd_card_helper)
{
  gboolean is_sd_present = FALSE;
  if (NULL != sd_card_helper->udev)
    return;
  sd_card_helper_debug ("start_udev_monitor");
  sd_card_helper->udev = udev_new ();
  if (!sd_card_helper->udev)
    {
      sd_card_helper_critical ("Unable to create udev object");
      return;
    }
  is_sd_present = is_sd_card_present (sd_card_helper->udev);
  if (is_sd_present)
    {
      sd_card_helper_debug ("+++SD is plugged in \n");
      com_bosch_pt_sd_card_helper_set_is_sd_card_present (
          sd_card_helper->sd_card_helper_object, TRUE);
    }
  else
    {
      sd_card_helper_debug ("---SD is not plugged in \n");
      com_bosch_pt_sd_card_helper_set_is_sd_card_present (
          sd_card_helper->sd_card_helper_object, FALSE);
    }

  sd_card_helper->udev_monitor
      = udev_monitor_new_from_netlink (sd_card_helper->udev, "udev");
  if (!sd_card_helper->udev_monitor)
    {
      sd_card_helper_critical ("Unable to create udev monitor");
      return;
    }

  if (udev_monitor_filter_add_match_subsystem_devtype (
          sd_card_helper->udev_monitor, UDEV_SUBSYSTEM_FILTER,
          UDEV_DEVTYPE_FILTER)
      < 0)
    {
      sd_card_helper_critical (
          "udev_monitor_filter_add_match_subsystem_devtype FAILED \n");
      return;
    }

  if (udev_monitor_enable_receiving (sd_card_helper->udev_monitor) < 0)
    {
      sd_card_helper_critical ("udev_monitor_enable_receiving FAILED \n");
      return;
    }

  sd_card_helper->udev_io_channel = g_io_channel_unix_new (
      udev_monitor_get_fd (sd_card_helper->udev_monitor));
  sd_card_helper->io_watch_udev
      = g_io_add_watch (sd_card_helper->udev_io_channel, G_IO_IN,
                        handle_udev_events, sd_card_helper);
}

/**
 * @author Vinay
 * @date 31.08.2020
 * @brief stop monitor on sd card added or removed.
 * @param sd_card_helper SDCardHelper pointer.
 */
static void
stop_udev_monitor (SDCardHelper *sd_card_helper)
{
  g_source_remove (sd_card_helper->io_watch_udev);
  g_io_channel_unref (sd_card_helper->udev_io_channel);
  udev_monitor_unref (sd_card_helper->udev_monitor);
  udev_unref (sd_card_helper->udev);

  sd_card_helper->io_watch_udev   = 0;
  sd_card_helper->udev_io_channel = NULL;
  sd_card_helper->udev_monitor    = NULL;
  sd_card_helper->udev            = NULL;
}

/**
 * @author Vinay
 * @date 31.08.2020
 * @brief check and notify sd card mount.
 * @param sd_card_helper SDCardHelper pointer.
 * @attention It will checked whether the added mount is sdcard or not and
 * it send notifiction that the sdcard mount path which is mounted.
 */
static void
check_and_notify_sd_card_mount (SDCardHelper *sd_card_helper)
{
  struct mntent *ent          = NULL;
  gboolean is_sd_card_mounted = FALSE;
  FILE *aFile                 = setmntent ("/proc/mounts", "r");
  if (aFile == NULL)
    {
      sd_card_helper_critical ("setmntent function call failed");
      return;
    }
  ent = getmntent (aFile);
  while (NULL != ent)
    {
      if (strncmp (ent->mnt_fsname, DEVICE_MMC, (strlen (DEVICE_MMC) - 1))
          == 0)
        {
          is_sd_card_mounted = TRUE;
          break;
        }
      ent = getmntent (aFile);
    }

  if (TRUE == is_sd_card_mounted && NULL == sd_card_helper->mount_path)
    {
      sd_card_helper->mount_path = g_strdup (ent->mnt_dir);
      sd_card_helper_debug ("SD card mounted to %s",
                            sd_card_helper->mount_path);
      com_bosch_pt_sd_card_helper_set_mount_point (
          sd_card_helper->sd_card_helper_object, sd_card_helper->mount_path);
    }
  else if (FALSE == is_sd_card_mounted && NULL != sd_card_helper->mount_path)
    {
      g_free (sd_card_helper->mount_path);
      sd_card_helper->mount_path = NULL;
      sd_card_helper_debug ("SD Mount Removed ");
      com_bosch_pt_sd_card_helper_set_mount_point (
          sd_card_helper->sd_card_helper_object, sd_card_helper->mount_path);
    }

  endmntent (aFile);
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Handler for 'mount-added' signal from GVolumeMonitor.
 * @param volume_monitor GVolumeMonitor pointer.
 * @param gmount         GMount pointer which represent the added mount.
 * @param user_data       pointer to user data
 * @attention It will checked whether the added mount is sdcard or not and
 * it send notifiction that the sdcard mount path which is mounted.
 */
static void
mount_added_callback (GVolumeMonitor *volume_monitor, GMount *gmount,
                      gpointer user_data)
{
  SDCardHelper *sd_card_helper = user_data;
  sd_card_helper_debug ("Mount Added Callback");
  check_and_notify_sd_card_mount (sd_card_helper);
}

/**
 * @author Vinay
 * @date 16.07.2020
 * @brief Handler for 'mount-removed' signal from GVolumeMonitor.
 * @param volume_monitor GVolumeMonitor pointer.
 * @param gmount         GMount pointer which represent the added mount.
 * @param user_data       pointer to user data
 * @attention We assume that umounted volume is sdcard, because there is no
 * other removable storage media in the device.
 */
static void
mount_removed_callback (GVolumeMonitor *volume_monitor, GMount *gmount,
                        gpointer user_data)
{
  SDCardHelper *sd_card_helper = user_data;
  sd_card_helper_debug ("Mount removed callback");
  check_and_notify_sd_card_mount (sd_card_helper);
}

/**
 * @author Vinay
 * @date 18.08.2020
 * @brief start monitor on sd card mount added or mount removed.
 */
static void
sd_card_mount_monitor (SDCardHelper *sd_card_helper)
{
  sd_card_helper_debug ("sd_crd_helper_start_mount_monitor");
  check_and_notify_sd_card_mount (sd_card_helper);
  sd_card_helper->volume = g_volume_monitor_get ();
  if (sd_card_helper->volume)
    {
      g_signal_connect (sd_card_helper->volume, "mount-added",
                        G_CALLBACK (mount_added_callback), sd_card_helper);
      g_signal_connect (sd_card_helper->volume, "mount-removed",
                        G_CALLBACK (mount_removed_callback), sd_card_helper);
    }
  else
    {
      sd_card_helper_critical ("Installing Volume Monitor failed");
    }
}

/**
 * @author Vinay
 * @date 11.03.2020
 * @brief Initilize fs and loop Plugin in libblockdev's function..
 */
static gint
blockdev_plugin_init ()
{
  gint ret;
  g_autoptr (GError) error = NULL;
  BDPluginSpec fs_plugin   = { BD_PLUGIN_FS, NULL };
  /*
   * The core daemon needs the part, swap, loop, mdraid, fs and crypto plugins.
   *   Additional plugins are required by various modules, but they make sure
   *   plugins are loaded themselves.
   */
  BDPluginSpec *plugins[] = { &fs_plugin, NULL, NULL };
  BDPluginSpec **plugin_p = NULL;
  error                   = NULL;

  /*
   * Skip runtime dependency checks when initializing libblockdev. Plugin
   *   shouldn't fail to load just because some if its dependencies is missing.
   */
  ret = bd_switch_init_checks (FALSE, &error);
  if (!ret)
    {
      sd_card_helper_error (
          "Error initializing libblockdev library: %s (%s, %d)",
          error->message, g_quark_to_string (error->domain), error->code);
      g_clear_error (&error);
    }

  ret = bd_try_init (plugins, NULL, NULL, &error);
  if (!ret)
    {
      if (error)
        {
          sd_card_helper_error (
              "Error initializing libblockdev library: %s (%s, %d)",
              error->message, g_quark_to_string (error->domain), error->code);
          g_clear_error (&error);
        }
      else
        {
          /*
           * a missing plugin is okay, calling functions from it will fail, but
           *   until that happens, life will just be great
           */
          for (plugin_p = plugins; *plugin_p; plugin_p++)
            if (!bd_is_plugin_available ((*plugin_p)->name))
              sd_card_helper_error (
                  "Failed to load the '%s' libblockdev plugin",
                  bd_get_plugin_name ((*plugin_p)->name));
        }
    }

  sd_card_helper_debug ("blockdev_plugin_init Done\n");

  return 0;
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief sd card helper init.
 */
void
sd_card_helper_init (SDCardHelper *sd_card_helper)
{
  sd_card_helper->mount_path = NULL;
  sd_card_helper_debug ("sd_card_helper_init");
  blockdev_plugin_init ();
  start_udev_monitor (sd_card_helper);
  sd_card_mount_monitor (sd_card_helper);
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief sd card helper deinit.
 */
void
sd_card_helper_deinit (SDCardHelper *sd_card_helper)
{
  sd_card_helper_debug ("sd_card_helper_deinit");

  /* Unmount the device if mounted before clean up */
  unmount_sd_card (sd_card_helper);

  stop_udev_monitor (sd_card_helper);
  if (sd_card_helper->volume)
    g_object_unref (sd_card_helper->volume);

  if (sd_card_helper->sd_card_helper_object)
    g_object_unref (sd_card_helper->sd_card_helper_object);
}
/**
 * @author Vinay
 * @date 27.02.2020
 * @brief Mount the Mass Storage Partition to Device.
 * @param dbus ComBosch_ptSdCardHelper
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 */
static gboolean
handle_mount (ComBosch_ptSdCardHelper *sd_card_helper_object,
              GDBusMethodInvocation *invocation, gpointer user_data)
{
  SDCardHelper *sd_card_helper = user_data;
  const gchar *mount_path      = SD_CARD_MOUNT_DIR;
  gchar *arg_mount_option      = NULL;
  gchar *mounted_path
      = mount_sd_card (sd_card_helper, mount_path, arg_mount_option);

  sd_card_helper_debug ("handle_mount\n");

  com_bosch_pt_sd_card_helper_complete_mount_sd_card (
      sd_card_helper_object, invocation, mounted_path);
  if (mounted_path)
    free (mounted_path);

  return TRUE;
}

/**
 * @author Vinay
 * @date 27.02.2020
 * @brief UnMount the Mass Storage Partition.
 * @param sd_card_helper_object ComBosch_ptSdCardHelper
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 */
static gboolean
handle_unmount (ComBosch_ptSdCardHelper *sd_card_helper_object,
                GDBusMethodInvocation *invocation, gpointer user_data)
{
  sd_card_helper_debug ("handle_unmount\n");
  SDCardHelper *sd_card_helper = user_data;
  unmount_sd_card (sd_card_helper);
  com_bosch_pt_sd_card_helper_complete_unmount_sd_card (sd_card_helper_object,
                                                        invocation);
  return TRUE;
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief freeing the task data
 * @param ptr data pointer.
 */
static void
format_fs_vfat_thread_data_free (gpointer ptr)
{
  SDCardHelper *data = ptr;
  free (data);
}

/**
 * @author Vinay
 * @date 11.06.2021
 * @brief callback in the current thread-default main context.
 * @param source_object the object the asynchronous operation was started with.
 * @param res a GAsyncResult.
 * @param user_data user data passed to the callback.
 */
static void
format_fs_vfat_thread_finish_cb (GObject *source_object, GAsyncResult *res,
                                 gpointer user_data)
{
  SDCardHelper *sd_card_helper           = user_data;
  g_autoptr (GError) error               = NULL;
  SdCardHelperFormatStatus format_status = SDCARD_FORMAT_STARTED;
  gssize ret = g_task_propagate_int (G_TASK (res), &error);
  sd_card_helper_debug ("Thread Return with ret = %d\n", ret);

  if (ret)
    {
      format_status = SDCARD_FORMAT_DONE;
    }
  else
    {
      format_status = SDCARD_FORMAT_FAILED;
    }

  com_bosch_pt_sd_card_helper_set_format_status (
      sd_card_helper->sd_card_helper_object, format_status);
  g_dbus_interface_skeleton_flush (
      (G_DBUS_INTERFACE_SKELETON (sd_card_helper->sd_card_helper_object)));
}

/**
 * @author Vinay
 * @date 24.11.2021
 * @brief Free the libfdisk context.
 * @param fdisk_context ctx pointer.
 */
static void
free_context (struct fdisk_context *cxt)
{
  fdisk_deassign_device (cxt, 0);
  fdisk_unref_context (cxt);
}

/**
 * @author Vinay
 * @date 24.11.2021
 * @brief Create the device partition if not present.
 * @param char device pointer.
 */
static void
create_partition_if_not_present (const char *device)
{
  int ret;

  if (access (device, R_OK | W_OK | F_OK) != 0)
    {
      sd_card_helper_critical ("Cannot access device %s \n", device);
      return;
    }

  struct fdisk_context *cxt = fdisk_new_context ();
  if (!cxt)
    {
      sd_card_helper_critical ("Failed to create fdisk context \n");
      return;
    }

  ret = fdisk_assign_device (cxt, device, 0);
  if (ret != 0)
    {
      sd_card_helper_critical ("Failed to assign device to fdisk context");
      fdisk_unref_context (cxt);
      return;
    }

  fdisk_disable_dialogs (cxt, 1);

  if (!fdisk_has_label (cxt))
    {
      ret = fdisk_create_disklabel (cxt, "dos");
      if (ret != 0)
        {
          sd_card_helper_critical ("Error while creating disk label");
          free_context (cxt);
          return;
        }

      ret = fdisk_write_disklabel (cxt);
      if (ret != 0)
        {
          sd_card_helper_critical ("Failed to write disk label");
          free_context (cxt);
          return;
        }

      ret = fdisk_reread_partition_table (cxt);
      if (ret != 0)
        {
          sd_card_helper_critical ("Failed to reread partition table");
          free_context (cxt);
          return;
        }
    }

  struct fdisk_table *table = fdisk_new_table ();
  if (!table)
    {
      sd_card_helper_critical ("fdisk_new_table call failed");
      free_context (cxt);
      return;
    }

  ret = fdisk_get_partitions (cxt, &table);
  if (ret != 0)
    {
      sd_card_helper_critical ("fdisk_get_partitions call failed");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  ret = fdisk_get_freespaces (cxt, &table);
  if (ret != 0)
    {
      sd_card_helper_critical ("fdisk_get_freespaces call failed");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  ret = fdisk_table_sort_partitions (table, fdisk_partition_cmp_start);
  if (ret != 0)
    {
      sd_card_helper_critical ("fdisk_table_sort_partitions call failed");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  struct fdisk_iter *itr = fdisk_new_iter (FDISK_ITER_FORWARD);
  if (!itr)
    {
      sd_card_helper_critical ("fdisk_new_iter call failed");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  gboolean part_present      = FALSE;
  struct fdisk_partition *pa = NULL;

  while (fdisk_table_next_partition (table, itr, &pa) == 0)
    {
      if (pa && !fdisk_partition_is_container (pa)
          && !fdisk_partition_is_freespace (pa))
        {
          sd_card_helper_debug ("Found partition in device");
          sd_card_helper_debug ("%s", fdisk_get_devname (cxt));
          sd_card_helper_debug ("no of partitions %d",
                                fdisk_partition_get_partno (pa) + 1);
          sd_card_helper_debug ("partition size :%llu ",
                                fdisk_partition_get_size (pa)
                                    * fdisk_get_sector_size (cxt));
          part_present = TRUE;
          break;
        }
    }

  fdisk_free_iter (itr);

  if (part_present)
    {
      sd_card_helper_debug ("Partition already present so not creating one");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  struct fdisk_partition *npa = fdisk_new_partition ();
  if (!npa)
    {
      sd_card_helper_critical ("Failed to create new partition");
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  fdisk_partition_start_follow_default (npa, 1);
  fdisk_partition_end_follow_default (npa, 1);
  fdisk_partition_partno_follow_default (npa, 1);

  ret = fdisk_add_partition (cxt, npa, NULL);
  if (ret != 0)
    {
      sd_card_helper_critical ("Failed to add new partition to table %d",
                               strerror (-ret));
      fdisk_unref_partition (npa);
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  ret = fdisk_write_disklabel (cxt);
  if (ret != 0)
    {
      sd_card_helper_critical ("Failed to write partition table to disk");
      fdisk_unref_partition (npa);
      fdisk_unref_table (table);
      free_context (cxt);
      return;
    }

  ret = fdisk_reread_partition_table (cxt);
  if (ret != 0)
    {
      sd_card_helper_critical ("Failed to reread the partition table");
    }

  fdisk_unref_partition (npa);
  fdisk_unref_table (table);
  free_context (cxt);
  return;
}
/**
 * @author Vinay
 * @date 11.06.2021
 * @brief task function to be run in a thread.
 * @param task the GTask
 * @param source task 's source object.
 * @param task_data task 's task data.
 * @param cancellable task 's GCancellable, or NULL.
 */
void
format_fs_vfat_thread (GTask *task, gpointer source, gpointer task_data,
                       GCancellable *cancellable)
{
  gint ret                 = 0;
  g_autoptr (GError) error = NULL;
  const gchar *device_name = DEVICE_MMC_PARTITION_NAME;
  gchar *fat_type          = "32";
  sd_card_helper_debug ("format_fs_vfat_thread Started\n");

  BDExtraArg bd_label_arg         = { "-n", "SD_CARD" };
  BDExtraArg bd_fat_type_arg      = { "-F", fat_type };
  const BDExtraArg *extra_args[3] = { &bd_label_arg, &bd_fat_type_arg, NULL };

  create_partition_if_not_present (DEVICE_MMC);

  ret = bd_fs_vfat_mkfs (device_name, extra_args, &error);

  if (!ret)
    {
      sd_card_helper_critical ("Error bd_fs_vfat_mkfs: %d\n", ret);
      sd_card_helper_critical ("%s (%s, %d)\n", error->message,
                               g_quark_to_string (error->domain), error->code);
    }
  else
    {
      sd_card_helper_debug ("bd_fs_vfat_mkfs Done: %d\n", ret);
    }
  g_task_return_int (task, ret);
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Format the Sd card Partition of the Device.
 * @param sd_card_helper ComBosch_ptSdCardHelper
 * @param invocation GDBusMethodInvocation - Object for handling remote calls.
 * @param user_data Data passed during handler installation. By Default NULL.
 * @return TRUE on success. FALSE otherwise.
 */
static gboolean
handle_format_fs_vfat (ComBosch_ptSdCardHelper *sd_card_helper_object,
                       GDBusMethodInvocation *invocation, gpointer user_data)
{
  SdCardHelperFormatStatus format_status = SDCARD_FORMAT_STARTED;
  SDCardHelper *data                     = NULL;
  GTask *task                            = NULL;
  sd_card_helper_debug ("handle_format_fs_vfat\n");
  com_bosch_pt_sd_card_helper_set_format_status (sd_card_helper_object,
                                                 format_status);
  g_dbus_interface_skeleton_flush (
      (G_DBUS_INTERFACE_SKELETON (sd_card_helper_object)));

  data = g_new0 (SDCardHelper, 1);
  task = g_task_new (NULL, NULL, format_fs_vfat_thread_finish_cb, user_data);
  g_task_set_task_data (task, data, format_fs_vfat_thread_data_free);

  /* Run the task in a worker thread and return immediately while that
   * continues in the background. When it’s done it will call @callback in
   * the current thread default main context. */
  g_task_run_in_thread (task, format_fs_vfat_thread);

  g_object_unref (task);

  com_bosch_pt_sd_card_helper_complete_format_vfat (sd_card_helper_object,
                                                    invocation);
  return TRUE;
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Invoked when a connection to D-bus has been obtained. Installs an
 * interface for an object, that can be remotely accessed via D-bus.
 * @param connection The GDBusConnection to a message bus.
 * @param name       The well known name of the application on the bus.
 * @param user_data  Data that was passed to g_bus_own_name().
 */
void
sd_card_helper_bus_acquired (GDBusConnection *connection, const gchar *name,
                             gpointer user_data)
{
  sd_card_helper_debug ("Acquired bus: %s\n", name);
  SDCardHelper *sd_card_helper = user_data;
  /*
   * Create new object
   * Creates a skeleton (server-) object instace for the D-Bus interface
   * "gdbus-interface-com.bosch_pt.SdCardHelper"
   */
  sd_card_helper->sd_card_helper_object
      = com_bosch_pt_sd_card_helper_skeleton_new ();

  /*
   * Install method/signal handler callbacks
   * Connect the callbacks to signal names for this server object.
   * The handler will be called before the default handler of the singal.
   *               ServerObject | "signal-name (Or) signal-name::detail" |
   *               Callback function |            Data to pass to callback.
   *               Default: NULL.
   */
  g_signal_connect (sd_card_helper->sd_card_helper_object,
                    "handle-unmount-sd-card", G_CALLBACK (handle_unmount),
                    sd_card_helper);
  g_signal_connect (sd_card_helper->sd_card_helper_object,
                    "handle-mount-sd-card", G_CALLBACK (handle_mount),
                    sd_card_helper);
  g_signal_connect (sd_card_helper->sd_card_helper_object,
                    "handle-format-vfat", G_CALLBACK (handle_format_fs_vfat),
                    sd_card_helper);

  /* Export the objects interface */
  if (!g_dbus_interface_skeleton_export (
          G_DBUS_INTERFACE_SKELETON (sd_card_helper->sd_card_helper_object),
          connection, "/com/bosch_pt/SdCardHelper", NULL))
    {
      sd_card_helper_critical (
          "Export ComBosch_ptSdCardHelper, interface skeleton failed\n");
      if (sd_card_helper->main_loop)
        g_main_loop_quit (sd_card_helper->main_loop);
    }
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Invoked when the name com_bosch_pt.SdCardHelper is acquired.
 *
 * @param connection The GDBusConnection on which to acquired to the name.
 * @param name       The name being owned.
 * @param user_data  Passed to g_bus_own_name() or
 * g_bus_own_name_no_connection().
 */
void
sd_card_helper_name_acquired (GDBusConnection *connection, const gchar *name,
                              gpointer user_data)
{
  sd_card_helper_debug ("Acquired the name: %s\n", name);
  SDCardHelper *sd_card_helper = user_data;

  sd_card_helper_init (sd_card_helper);
}

/**
 * @author Vinay
 * @date 16.08.2020
 * @brief Invoked when the name com.bosch_pt.SdCardHelper is lost or connection
 * has been closed.
 *
 * @param connection The GDBusConnection on which to acquire the name or NULL
 * if the connection was disconnection.
 * @param name       The name being owned.
 * @param user_data  Passed to g_bus_own_name() or
 * g_bus_own_name_no_connection().
 */
void
sd_card_helper_name_lost (GDBusConnection *connection, const gchar *name,
                          gpointer user_data)
{
  sd_card_helper_critical ("Lost the name: %s\n", name);
  SDCardHelper *sd_card_helper = user_data;
  /* quit GMainLoop. */
  if (sd_card_helper->main_loop)
    g_main_loop_quit (sd_card_helper->main_loop);
}
